/*
   Program: answer.prg
   Written by:  Charles Nathan of Sun Tzu, Inc.  dba Pinnacle Development
   Changes:
       08-19-93 CN original version

   Demonstration program for a simple two-line answering machine written
   using the Clipper Interface Library for Dialogic.

   The answering machine provides for recording and retreival of messages
   to specific mailboxes or a general mailbox.  Persons with a mailbox
   may retreive their messages or a general message after entering a
   password.  The program will attempt to play a specific message depending
   on whether the call is on a weekend or holiday, at lunch or outside of
   normal business hours.

   Coments throughout the program refer to a person calling in to get
   their messages as a super-user.  The way a user accesses their messages
   is to press the "#" key at the box selection menu.  (the first menu
   of choices)  Then they enter a three digit password which is stored
   in the file PASSWORD.DBF.

   The way the terminating characters are set up, this program will only
   work for less than ten boxes.  By changing the max number of digits and
   to a # key terminating character, multi-digit box numbers can be used.

   When no calls are being processed, the F2 key will show current messages.

   If there are messages waiting, the program makes an inoccous ticking
   noise until all messages are reviewed and deleted.  This can be turned
   off with the NOTIFY manifest constant by making it .F.

   REQUIRED FILES
   PASSWORD.DBF      System passwords for super-users
   MESSANS.DBF       Messages, storing date and time, destination box and associated VOX file
   HOLIDAY.VOX       Message to say when its a weekend or a holiday
   TIMING.VOX        Message to say when call is outside of defined business hours
   LUNCH.VOX         Message to say when call is during lunch
   BUSY.VOX          Message to say when call is during normal business hours, but not answered
   NAMES.VOX         Plays all of the mailboxes and their corresponding numbers
   PETER.VOX         Personal messages
   CRAIG.VOX         ...
   CHARLES.VOX
   BOB.VOX
   GENERAL.VOX      General message available for everyone.
   MESSx.VOX        Recorded messages.
   HELPMESS.VOX     The message giving you your options after you record.
   PASSWORD.VOX     Prompt for entry of password
   USERLIST.VOX     List of users
   USERMENU.VOX     Menu options for a super-user
   MOPTIONS.VOX     Messsage options for review/delete of a message.

   In addition, the following index numbers in GLOBAL.VAP are expected to be:

   34        "AM"
   35        "PM"
   400       "message"
   401       "was received on"
   402       "at"
   403       "will be deleted when you hang up"
   404       "will be saved when you hang up"
   405       "you have"
   406       "new messages"
   407       "no"
   408       "new message"

   This code makes reference to the "INITIATE" function and the "PROCESS"
   function.  An INITIATE and PROCESS function are always set up in
   pairs.  First an INITIATE function will start some activity, say
   putting a channel's phone line on-hook.  (hanging it up)  Then the
   PROCESS function will check the status of the activity, and set up
   the next pair of activities.

   In this way, the flow of processing is always flexible.  Depending
   on the result of an action created by a INITIATE function, the
   PROCESS function determines what to do next.

   NOTE:  The Dialogic board hardware interrupt is defined as a manifest
          constant, HWIRQ.  Make sure you change this to be the same
          as the hardware interrupt for your Dialogic board before you
          complile pdmenu.prg to try it, or always pass the irq to the
          program on the command line.

   Compile pdmenu by:

     CLIPPER answer /m /n /q /v /a /w /b

   This will include debugging trace information so that you can single-step
   through the program with the Clipper debugger and get a better idea of
   the flow.  Use the "code block trace" option ON so that you can see where
   code blocks were created as they are being executed.

   The DEBUGDELAY manifest constant is set to a small increment of time.
   In the state machine, if you use INKEY() rather than INKEY(nSmall),
   the state machine won't stop long enough for you to pop-up to debugger.
   DEBUGDELAY gives it enough time to see an "Alt-D" without affecting
   the performance of the system.  For a final version, you could remove
   the delay.

*/

#INCLUDE "PDDIAL.CH"
#INCLUDE "INKEY.CH"
#INCLUDE "SETCURS.CH"

#define  HWIRQ         5                         // default hardware interrupt
#define  NUM_RINGS     3                         // number of rings before answering
#define  DEBUGDELAY    .05                       // delay to see Alt-D, see above
#define  INITIATE      1                         // array element for initiate function
#define  PROCESS       2                         // array element for process fuction
#define  EVENT         -1                        // -1 if an event occured

#define K_MESSAGES     K_F2                      // hotkey for message view

// manifest constants for normal business and lunch hours
#define SHOP_OPEN  '08:30'
#define SHOP_CLOSE '18:00'
#define SHOP_LUNCH '12:00'
#define SHOP_BACK  '13:00'

#define MAX_RECORDING   120                      // maximum recording time: 120 secs (2 min)
#define MAX_REC_SILENCE 10                       // max silence interval: 5 sec

#define NOTIFY          .T.                      // should the thing click away if there are messages waiting?

// file-wide static variables.  These are visible anywhere in this source
// module.
STATIC   nEvent,;                                // the last event number that occured
         nChan,;                                 // the last channel number that an event occured on
         aCurBlock,;                             // the current (next to be executed) initiate and process function code blocks
         aFileHandle,;                           // an array to store the file handle number of a VOX or VAP file being played.  Each array element corresponds to a channel
         aRecFile,;                              // file being recorded into
         aPerson,;                               // array of person names, channel by channel
         aNewMess,;                              // array of new messages for each channel
         aCurMess,;                              // array of current messages for each channel
         aHolidays                               // array of legal holiday dates

//
FUNCTION Main(cHWIRQ)
// this is the main function which contains the "state machine" loop.
// the state machine repetitively checks for events which have occured,
// and processes them one by one.   The hardware interrupt number can be
// passed from DOS as a command line parameter.

LOCAL nHWIRQ,;                                   // passed or default hardware interrupt of the Dialogic board as a numeric
      nInkey,;                                   // keycode last pressed
      nChannel,;                                 // channel number loop counter for channel initialization
      nErr,;                                     // error status returned from initiate function
      nMaxChan                                   // maximum channels available


IF EMPTY(cHWIRQ)
   nHWIRQ := HWIRQ
ELSE
   nHWIRQ := VAL(cHWIRQ)
ENDIF

CLEAR SCREEN
SETCURSOR(SC_NONE)

winMsg('Initialize on IRQ ' + ALLTRIM(STR(nHWIRQ)) )

IF (nErr:= pd_Init(nHWIRQ, @nMaxChan,;
                  C_LC + C_RING + C_OFFH + C_ONH, NUM_RINGS)) == 0


   // initialize static channel data arrays
   aRecFile    := AFILL( ARRAY(nMaxChan), '')         // .VOX file to record into
   aCurBlock   := AFILL( ARRAY(nMaxChan), { {|| NIL}, {|| NIL} } )  // next pair of code blocks to execute
   aFileHandle := AFILL( ARRAY(nMaxChan), 0)          // .VOX file being played
   aPerson     := AFILL( ARRAY(nMaxChan), '')         // person currently on-line
   aNewMess    := AFILL( ARRAY(nMaxChan), 0)          // new message count
   aCurMess    := AFILL( ARRAY(nMaxChan), 0)          // current message number

   // load an array of all of the legal holidays this year
   WinMsg('Loading holiday list')
   aHolidays   := LegalHolidays(DATE())

   FOR nChannel := 1 to nMaxChan
      // set up the first event, go on-hook on each channel
      aCurBlock[nChannel] := { { || OnHookInit() }, { || OnHookProc() } }
   NEXT

   // set initial values of global parameters
   pd_global('loopsig',    1)                 // set to terminate with a change in loop current (hang-up)
   pd_global('termdtmf', '@')                 // set to terminate touch-tone entry with any character pressed

   // open data files
   USE PASSWORD EXCLUSIVE NEW                 // storage of user passwords
   USE MESSANS  EXCLUSIVE NEW                 // store of messages



   DO WHILE (nInkey := INKEY(DEBUGDELAY)) <> K_ESC
      // If the hotkey to view messages is pressed, go do it.
      // This should not be done while calls are being processed because
      // it stops the state machine!!
      IF nInKey == K_MESSAGES
         ViewMessages()
      ENDIF

      IF pd_GetEvent(@nChan, @nEvent) == EVENT
         WinMsg('Processing calls')
         // pd_getEvent() returns a -1 if an event has occured
         // if an event occurs, evaluate the PROCESS function
         EVAL( aCurBlock[nChan,PROCESS] )

         // evaluate the INITIATE function for the next activity
         IF (nErr:= EVAL (aCurBlock[nChan,INITIATE])) <> 0
            // an error occured, display it.
            pd_Error(nErr)
         ENDIF
      ELSE
         // if there is no event, beep to let us know there are
         // messages waiting.
         IF NOTIFY .AND. MESSANS->(RECCOUNT()) <> 0
            TONE(20000, .1)
         ENDIF
      ENDIF
   ENDDO

   // the state machine has terminated, close our files
   CLOSE MESSANS
   CLOSE PASSWORD

   // Shut down the voice system
   IF pd_deinit(nMaxChan) <> 0
      WinMsg( ALLTRIM(pd_d4xerr(nErr)) )
   ENDIF
ELSE

   // initialization failed, display the error message
   WinMsg(  ALLTRIM(pd_d4xerr(nErr)) )
ENDIF

CLEAR SCREEN

RETURN NIL

//
FUNCTION OnHookInit()
// place the current line on-hook
LOCAL nStatus

pd_DispMsg('OnHookInit')
// note that each of these functions displays a little message to the screen
// so you can see what's happening.  This is a good idea to do in your program
// too, especially for debugging!  When you are ready to ship, just make a
// function like:
// FUNCTION pd_DispMsg()
// RETURN NIL
// then your messages will no longer be written to the screen.  As an
// excercise left for the user, have your pd_dispMsg() function write
// each activity to a file!  Wouldn't that make a nice log of what had
// happened?

nStatus := pd_SetHook(nChan, H_ONH)

RETURN nStatus

//
FUNCTION OnHookProc()
// set up the next activity to process a ring on the line

pd_DispMsg('OnHookProc')
aCurBlock[nChan] := {    { || RingInit() }, { || RingProc() }    }

// it is not necessary for a PROCESS function to return anything

RETURN NIL

//
FUNCTION RingInit()

pd_DispMsg('RingInit')
// clear any pending DTMF digits on the channel
pd_ClrDtmf(nChan)

// of course, an INITIALIZE function must return an approprate numeric status
RETURN E_SUCC

//
FUNCTION RingProc()
// process a ring on the line, go off-hook.

pd_DispMsg('RingProc')
IF nEvent == T_RING
   aCurBlock[nChan] := {    { || OffHookInit() },  { || OffHookProc() }   }
ELSE
   // if it wasn't a ring, which is what we were expecting, go back on hook.
   // (try to reset the line status to a known state)
   aCurBlock[nChan] := {    { || OnHookInit() },  { || OnHookProc() }   }
ENDIF

RETURN NIL

//
FUNCTION OffHookInit()
LOCAL nStatus

pd_DispMsg('OffHookInit')

// change the channel to off-hook.  See pd_SetHook for possible values
// of the second parameter.  (USE MANIFEST CONSTANTS rather than numeric
// values to make your code more readable)

nStatus := pd_SetHook(nChan, H_OFFH)

RETURN nStatus


//
FUNCTION OffHookProc()
// set up the next activity to play an introductory message

pd_DispMsg('OffHookProc')

aCurBlock[nChan] := {    { || PlayInit() },  { || PlayProc() }   }

RETURN NIL

//
FUNCTION PlayInit()
// play the machine message approprate to the kind of day it is.
LOCAL dDate,;                                    // the current date
      cTime,;                                    // the current time
      nStatus,;
      nHandle  := 0

pd_DispMsg('PlayInit')
dDate := Date()
cTime := LEFT(TIME(), 5)

IF CDOW(dDate) == 'Saturday' .OR.;
   CDOW(dDate) == 'Sunday'   .OR.;
   Holiday(dDate, aHolidays)
   // if its a saturday, sunday or legal holiday, play an appropriate message,
   // something like:  our office is closed today, please leave a message.
   nStatus := pd_Play(nChan, 'HOLIDAY.VOX', @nHandle)
   pd_DispMsg('PlayInit HOLIDAY')

ELSEIF cTime <= SHOP_OPEN .OR. cTime >= SHOP_CLOSE
   // outside normal business hours, play a closed for the day
   nStatus := pd_Play(nChan, 'TIMING.VOX', @nHandle)
   pd_DispMsg('PlayInit CLOSED')

ELSEIF cTime >= SHOP_LUNCH .AND. cTime <= SHOP_BACK
      // lunch hour
   nStatus := pd_Play(nChan, 'LUNCH.VOX', @nHandle)
   pd_DispMsg('PlayInit LUNCH')

ELSE
   // can't come to the phone right now.
   nStatus := pd_Play(nChan, 'BUSY.VOX', @nHandle)
   pd_DispMsg('PlayInit BUSY')

ENDIF

aFileHandle[nChan] := nHandle

RETURN nStatus

//
FUNCTION PlayProc()
// set up next activity to prompt for the mailbox name

pd_DispMsg('PlayProc')

IF nEvent == T_EOF .OR. nEvent == T_TERMDT
   // we expect an EOF or digit terminating the play of the last message
   aCurBlock[nChan] :=  {    { || NameInit() },  { || NameProc() }   }
ELSE
   // if not what was expected, go back on-hook.
   aCurBlock[nChan] :=  {    { || OnHookInit() },  { || OnHookProc() }   }
ENDIF

// close the file previously played
pd_VOXClose( aFileHandle[nChan] )

RETURN NIL

//
FUNCTION NameInit()
// play the file with information about the available mailboxes.
LOCAL nStatus,;                                  // status of the play
      nHandle                                    // save the handle of NAMES.VOX for later closing

pd_DispMsg('NameInit')
nStatus := pd_Play(nChan, 'NAMES.VOX', @nHandle)

aFileHandle[nChan] := nHandle

RETURN nStatus

//
FUNCTION NameProc()
// check for sucessful play of NAMES.VOX.  Set up the next activity to play
// the instructions VOX file.

pd_DispMsg('NameProc')

IF nEvent == T_EOF .OR. nEvent == T_TERMDT
   // we expect an EOF or digit terminating the play of the last message
   // Setup to next get input regarding the destination voice mailbox.
   aCurBlock[nChan] := {    { || BoxSelectInit() },  { || BoxSelectProc() }   }
ELSE
   // if not what was expected, go back on-hook.
   aCurBlock[nChan] := {    { || OnHookInit() },  { || OnHookProc() }   }
ENDIF

// close NAMES.VOX
pd_VOXClose( aFileHandle[nChan] )

RETURN NIL

//
FUNCTION BoxSelectInit()
// get entered DTMF digits in response to "which mailbox?" question
LOCAL nStatus

pd_DispMsg('BoxSelectInit')

// retrieve entered DTMF digits.
pd_global('termdtmf', '@')                       // set to terminate on any digit
pd_global('maxsil',     5)                       // maximum seconds, replay options
pd_global('loopsig',    1)                       // terminate on drop of line
nStatus := pd_getDTMFs(nChan)

RETURN nStatus


//
FUNCTION BoxSelectProc()
// process entry for the selection of destination voice mailbox.

LOCAL  cBufContent,;
       nStatus,;
       nUser,;
       nHandle

pd_DispMsg('BoxSelectProc')
IF nEvent == T_TERMDT
   // they pressed _SOMETHING_
   cBufContent := pd_getBuf(nChan)
   nUser       := VAL(cBufContent)

   // a number between 0 and 10, play their personal message
   IF nUser > 0 .AND. nUser < 10
      // locate password file to the proper user, index is record number
      PASSWORD->(DBGOTO(nUser))
      // set static for person's name
      aPerson[nChan]   := RTRIM(PASSWORD->name)
      // set next activity to play the person's personal message
      aCurBlock[nChan] := {    { || PlayMessInit(RTRIM(PASSWORD->name) + '.VOX') },  { || PlayMessProc() }   }

   ELSEIF cBufContent == '#'
      // prompt for password to see if this is a super-user
      aCurBlock[nChan] := {    { || PlayIDInit() },  { || PlayIDProc() }   }

   ELSE
      // dunno, go back on-hook.
      aCurBlock[nChan] := {    { || OnHookInit() },  { || OnHookProc() }   }
   ENDIF

ELSEIF nEvent == T_LCTERM
   // They hung-up.  Do the same to them.
   aCurBlock[nChan] := {    { || OnHookInit() },  { || OnHookProc() }   }

ELSEIF nEvent == T_SIL
    // replay the message, maybe they fell asleep...
    aCurBlock[nChan] := {    { || NameInit() },  { || NameProc() }   }

ELSE
   // unknown state, go back on-hook
   aCurBlock[nChan] := {    { || OnHookInit() },  { || OnHookProc() }   }
ENDIF

RETURN NIL

//
FUNCTION PlayMessInit(cFile)
// play the personal message for the selected person.

LOCAL nStatus,;
      nHandle

pd_DispMsg('PlayMessInit')

IF FILE(cFile)
   nStatus := pd_Play(nChan, cFile, @nHandle)
ELSE
   // if the person does not have a personal message, play the general one.
   nStatus := pd_Play(nChan, 'GENERAL.VOX', @nHandle)
ENDIF

aFileHandle[nChan] := nHandle

RETURN nStatus

//
FUNCTION PLayMessProc()
// check for the sucessful play of the personal message.  Next activity
// is to record the person's message.

pd_DispMsg('PlayMessProc')
IF nEvent == T_EOF .OR. nEvent == T_TERMDT
   // a sucessful termination, set up next to record.
   aCurBlock[nChan] := {    { || RecInit() },  { || RecProc() }   }
ELSE
   aCurBlock[nChan] := {    { || OnHookInit() },  { || OnHookProc() }   }
ENDIF

pd_VOXClose( aFileHandle[nChan] )

RETURN NIL

//
FUNCTION RecInit()
LOCAL  nNextPos,;
       nStatus,;
       nHandle

pd_DispMsg('RecInit')
// figure out the next message sequence number
MESSANS->(DBGOBOTTOM())
nNextPos := VAL( RTRIM( RIGHT(MESSANS->message_id, 4))) + 1

MESSANS->(DBAPPEND())
REPLACE MESSANS->message_id WITH 'MESS' + LTRIM(STR(nNextPos)),;
        MESSANS->person     WITH aPerson[nChan],;
        MESSANS->date       WITH Date(),;
        MESSANS->time       WITH MILAMPM( TIME() ),;
        MESSANS->status     WITH 'NEW'

aRecFile[nChan] := RTRIM(MESSANS->message_id) + '.VOX'

// reset some global stuff optimally for recording
pd_clrdtmf(nChan)                                // clear any pending DTMF digits

// setup recording parameters, see function elsewhere in this module
RecordSetup()

// record with silence compression, don't append to the file if it exists.
nStatus := pd_RecFile(nChan, aRecFile[nChan], @nHandle, RM_SCOMP, .F.)

aFileHandle[nChan] := nHandle

RETURN nStatus

//
FUNCTION RecProc()
// check status of the message record

pd_DispMsg('RecProc')

IF nEvent == T_EOF .OR. nEvent == T_TERMDT
   // next activity is to play the help menu to check your recording
   aCurBlock[nChan] := {    { || HelpMessInit() },  { || HelpMessProc() }   }
ELSE
   // unknown state, go back on-hook.
   aCurBlock[nChan] := {    { || OnHookInit() },  { || OnHookProc() }   }
ENDIF

// close the file we just recorded
pd_VOXClose( aFileHandle[nChan] )

// reset record/play parameters
pd_reset()
pd_global('loopsig', 1)

RETURN NIL

//
FUNCTION HelpMessInit()
// intialize play of the help message to modify a message.

LOCAL nStatus,;
      nHandle

pd_DispMsg('HelpMessInit')

pd_ClrDTMFs(nChan)

pd_global('termdtmf', '@')
nStatus := pd_Play(nChan, 'HELPMESS.VOX', @nHandle)
aFileHandle[nChan] := nHandle

RETURN nStatus

//
FUNCTION HelpMessProc()
// check the play of the help message.

pd_DispMsg('HelpMessProc')
IF nEvent == T_EOF .OR. nEvent == T_TERMDT
   // successful play, go next to process input for review activities
   aCurBlock[nChan] := {    { || ReviewInit() },  { || ReviewProc() }   }
ELSE
   aCurBlock[nChan] := {    { || OnHookInit() },  { || OnHookProc() }   }
ENDIF

// close the help message file
pd_VOXClose( aFileHandle[nChan] )

RETURN NIL

//
FUNCTION ReviewInit()
// review a message
LOCAL nStatus

pd_DispMsg('ReviewInit')

// set to terminate touch-tone entry with any character pressed
pd_global('termdtmf', '@')
nStatus := pd_getDTMFs(nChan)

RETURN nStatus

//
FUNCTION ReviewProc()
LOCAL  cBufContent

pd_DispMsg('ReviewProc')

IF nEvent == T_TERMDT
   cBufContent := RemTermDTMF( pd_getBuf(nChan) )
   DO CASE
      CASE cBufContent == '1'
         // play back the message
         aCurBlock[nChan] := {    { || PlayBackInit() },  { || PlayBackProc() }   }

      CASE cBufContent == '2'
         // add to the message
         aCurBlock[nChan] := {    { || AppendInit() },  { || AppendProc() }   }

      CASE cBufContent == '3'
         // re-record the message
         aCurBlock[nChan] := {    { || ReRecInit() },  { || ReRecProc() }   }

      CASE cBufContent == '4'
         //
         aCurBlock[nChan] := {    { || NameInit() },  { || NameProc() }   }

      CASE cBufContent == '0'
         // hang up
         aCurBlock[nChan] := {    { || OnHookInit() },  { || OnHookProc() }   }

      OTHERWISE
         // play the help message again
         aCurBlock[nChan] := {    { || HelpMessInit() },  { || HelpMessProc() }   }
   ENDCASE
ELSE
   // unknown state, hang up the line
   aCurBlock[nChan] := {    { || OnHookInit() },  { || OnHookProc() }   }
ENDIF

RETURN NIL

//
FUNCTION PlayBackInit()

// play back the message you recorded
LOCAL nStatus,;
      nHandle

pd_DispMsg('PlayBackInit')

nStatus := pd_Play(nChan, aRecFile[nChan], @nHandle)
aFileHandle[nChan] := nHandle

RETURN nStatus

//
FUNCTION PlayBackProc()
// check for sucessful play of the message

pd_DispMsg('PlayBackProc')

IF nEvent == T_EOF .OR. nEvent == T_TERMDT
   // play the menu again
   aCurBlock[nChan] := {    { || HelpMessInit() },  { || HelpMessProc() }   }
ELSE
   aCurBlock[nChan] := {    { || OnHookInit() },  { || OnHookProc() }   }
ENDIF

// close the message VOX file
pd_VOXClose( aFileHandle[nChan] )

RETURN NIL

//
FUNCTION AppendInit()
// add to the message you already recorded
LOCAL nStatus,;
      nHandle

pd_DispMsg('AppendInit')

// setup parameters for recording
RecordSetup()

nStatus := pd_RecFile(nChan, aRecFile[nChan], @nHandle, RM_SCOMP, .T.)
aFileHandle[nChan] := nHandle

RETURN nStatus

//
FUNCTION AppendProc()
// check for sucessful record of the message (appending)
pd_DispMsg('AppendProc')

IF nEvent == T_EOF .OR. nEvent == T_TERMDT
   // play the menu again
   aCurBlock[nChan] := {    { || HelpMessInit() },  { || HelpMessProc() }   }
ELSE
   // hang up
   aCurBlock[nChan] := {    { || OnHookInit() },  { || OnHookProc() }   }
ENDIF

// close the message file
pd_VOXClose( aFileHandle[nChan] )

// reset to normal playback parameters
pd_reset()
pd_global('loopsig', 1)

RETURN NIL

//
FUNCTION ReRecInit()
// re-record a message
LOCAL nStatus,;
      nHandle

pd_DispMsg('ReRecInit')

RecordSetup()
nStatus := pd_RecFile(nChan, aRecFile[nChan], @nHandle, RM_SCOMP, .F.)

aFileHandle[nChan] := nHandle

RETURN nStatus

//
FUNCTION ReRecProc()
// check for success of message re-record

pd_DispMsg('ReRecProc')

IF nEvent == T_EOF .OR. nEvent == T_TERMDT
   aCurBlock[nChan] := {    { || HelpMessInit() },  { || HelpMessProc() }   }
ELSE
   aCurBlock[nChan] := {    { || OnHookInit() },  { || OnHookProc() }   }
ENDIF

// close the message file
pd_VOXClose( aFileHandle[nChan] )

pd_reset()
pd_global('loopsig', 1)

RETURN NIL

//
FUNCTION PlayIDInit()
// play prompt to enter your password

LOCAL  nHandle,;
       nEvent

pd_DispMsg('PlayIDInit')

// " please enter a 3 digit password followed by the pound sign"
nEvent := pd_Play(nChan, 'PASSWORD.VOX', @nHandle)
aFileHandle[nChan] := nHandle

// re-set some global parameters
// reset terminating character to '#' to allow more than one digit to be entered
pd_global('termdtmf', '#')
pd_global('maxsil',    20)
RETURN nEvent

//
FUNCTION PlayIDProc()
// check for success of enter password prompt

pd_DispMsg('PlayIDProc')

IF nEvent == T_TERMDT .OR. nEvent == T_EOF
   aCurBlock[nChan] :=  {    { || CheckIDInit() },  { || CheckIDProc() }   }
ELSE
   aCurBlock[nChan] := {    { || PlayInit() },  { || PlayProc() }   }
ENDIF

// close the "enter password" VOX file
pd_VOXClose( aFileHandle[nChan] )

RETURN NIL

//
FUNCTION CheckIDInit()
// set up for entry of password
LOCAL nStatus

pd_DispMsg('CheckIDInit')

pd_global('termdtmf', '#')
nStatus := pd_getDTMFs(nChan)


RETURN nStatus

//
FUNCTION CheckIDproc()
// verify entered password
LOCAL cBufContent

pd_DispMsg('CheckIDProc')

IF nEvent == T_TERMDT
   cBufContent := RemTermDTMF( pd_GetBuf(nChan) )

   IF EMPTY(cBufContent)
      // they are trying to get out of this menu, back to the menu
      // for selecting who to leave a message to.
      aCurBlock[nChan] := {    { || NameInit() },  { || NameProc() }   }

   ELSE
      // try to find them in the user file
      // this would be nicer with a SEEK, so do it if you want.
      SELECT PASSWORD
      LOCATE FOR cBufContent == PASSWORD->id

      IF PASSWORD->( FOUND() )
         // set static value for who's on line
         aPerson[nChan]   := PASSWORD->name
         // set up to play new messages
         aCurBlock[nChan] := {    { || NewMessInit() },  { || NewMessProc() }   }
      ELSE
         // prompt for password again.
         aCurBlock[nChan] := {    { || PlayIDInit() },  { || PlayIDProc() }   }
      ENDIF
   ENDIF

ELSE
   // unknown state, hang up.
   aCurBlock[nChan] := {    { || OnHookInit() },  { || OnHookProc() }   }
ENDIF

// set global terminating character back to any digit terminates
pd_global('termdtmf',    '@')

RETURN NIL

//
FUNCTION NewMessInit()
// play new messages notification
LOCAL nGlobalHandle,;                            // handle of the global file
      nStatus,;
      aPlay := {}

pd_DispMsg('NewMessInit')

nGlobalHandle := pd_handGlobal()

// count the new messages
aNewMess[nChan] := 0                             // reset new message counter to zero

MESSANS->(DBGOTOP())
DO WHILE .NOT. MESSANS->(EOF())

   // It also checks to see if there is any General messages
   IF (MESSANS->person = aPerson[nChan] .AND. MESSANS->status = "NEW") .OR. ;
                      MESSANS->person = "GENERAL"
      aNewMess[nChan]++
   ENDIF
   MESSANS->(DBSKIP())
ENDDO

// build an array of messages to play out of GLOBAL.VAP.
AADD(aPlay, {nGlobalHandle, 405} )                 // "You have"

IF aNewMess[nChan] == 0
   AADD(aPlay, {nGlobalHandle, 407})               // "no"
ELSE
   // assumes not more than nine messages
   AADD(aPlay, {nGlobalHandle, aNewMess[nChan]})   // <number>
ENDIF

IF aNewMess[nChan] == 1
   AADD(aPlay, {nGlobalHandle, 408} )            // "new message"
ELSE
   AADD(aPlay, {nGlobalHandle, 406} )            // "new messages"
ENDIF

// play "you have n new messages"
nStatus := pd_PlaySeq(nChan, aPlay)

RETURN nStatus

//
FUNCTION NewMessProc()
// check for sucessful completion of the new messages anouncement

pd_DispMsg('NewMessProc')

IF nEvent == T_EOF .OR. nEvent == T_TERMDT
   // if a normal termination of the play, go to the super-user menu
   aCurBlock[nChan] := {    { || StartMenuInit() },  { || StartMenuProc() }   }
ELSE
   // unknown state, hang up.
   aCurBlock[nChan] := {    { || OnHookInit() },  { || OnHookProc() }   }
ENDIF

RETURN NIL

//
FUNCTION StartMenuInit()
LOCAL nStatus,;
      nHandle

pd_DispMsg('StartMenuInit')
IF aNewMess[nChan] == 0
   // Press 2 to leave a message or any other key to hang up
   // should imply that you don't have any messages to hear.
   nStatus := pd_Play(nChan, 'USERLIST.VOX', @nHandle)
ELSE
   // Press 1 to listen to your messages, 2 to leave a message etc.
   nStatus := pd_Play(nChan, 'USERMENU.VOX', @nHandle)
ENDIF

aFileHandle[nChan] := nHandle

RETURN nStatus

//
FUNCTION StartMenuProc()

pd_DispMsg('StartMenuProc')

IF nEvent == T_EOF .OR. nEvent == T_TERMDT
   // annouce the user menu options
   aCurBlock[nChan] := {    { || UserMenuInit() },  { || UserMenuProc() }   }
ELSE
   // unknown state, hang up.
   aCurBlock[nChan] := {    { || OnHookInit() },  { || OnHookProc() }   }
ENDIF

pd_VOXClose( aFileHandle[nChan] )

RETURN NIL

//
FUNCTION UserMenuInit()
// set up for DTMF input to the user menu

LOCAL nStatus

pd_DispMsg('UserMenuInit')

pd_global('termdtmf', '@')
nStatus := pd_getDTMFs(nChan)

RETURN nStatus

//
FUNCTION UserMenuProc()
// analyze input for the user menu
LOCAL cBufContent

pd_DispMsg('UserMenuProc')

IF nEvent == T_TERMDT
   cBufContent := pd_getBuf(nChan)
   DO CASE
      CASE cBufContent == '1'
         // listen to one of your messages
         IF aNewMess[nChan] > 0
            // reset the current message number to one
            aCurMess[nChan]  := 1
            aCurBlock[nChan] := {    { || messInfoInit() },  { || messInfoProc() }   }
         ELSE
            // no new messages, hang up.
            aCurBlock[nChan] := {    { || OnHookInit() },  { || OnHookProc() }   }
         ENDIF

      CASE cBufContent == '2'
         // leave a message for someone else
         aCurBlock[nChan] := {    { || NameInit() },  { || NameProc() }   }

      OTHERWISE
         // hang-up
         aCurBlock[nChan] := {    { || OnHookInit() },  { || OnHookProc() }   }
   ENDCASE

ELSE
   // unknown state, hang up.
   aCurBlock[nChan] := {    { || OnHookInit() },  { || OnHookProc() }   }
ENDIF

RETURN NIL

//
FUNCTION messInfoInit()
// play information about the message, when it was received, etc.
LOCAL nGlobalHandle,;                            // handle of the global prompt file
      nStatus,;                                  // return status for the play init
      dDate,;                                    // message date
      cTime,;                                    // message time
      aPlay := {},;                              // array of voice snippets to play
      nHour,;                                    // hour of message
      nMinute,;                                  // minute of message
      cAMPM,;                                    // AM PM of message
      nRecNo

pd_DispMsg('messInfoInit')

nGlobalHandle := pd_handGlobal()

nRecNo:=FindRecord()

MESSANS->(DBGOTO(nRecNo))

dDate           := MESSANS->date
cTime           := MESSANS->time
aRecFile[nChan] := RTRIM(MESSANS->message_id) + '.VOX'

AADD(aPlay, {nGlobalHandle, 400})                // "Message"
AADD(aPlay, {nGlobalHandle, aCurMess[nChan]})    //  <number>
AADD(aPlay, {nGlobalHandle, 401})                //  "was received on"
AADD(aPlay, {nGlobalHandle, 100+DOW(dDate)})     //  <day of the week>
AADD(aPlay, {nGlobalHandle, 200+MONTH(dDate) })  //  <name of the month>
AADD(aPlay, {nGlobalHandle, 300+DAY(dDate) })    //  <day>
AADD(aPlay, {nGlobalHandle, 402})                //  "at"

nHour   := VAL( LEFT(cTime, 2) )
nMinute := VAL( LEFT( RIGHT(cTime, 3), 2 ) )

// add the equivelent numbers to the aPlay array for the hour and the minute
pd_numSeq(nGlobalHandle, nHour,   aPlay)
IF nMinute <> 0
   pd_numSeq(nGlobalHandle, nMinute, aPlay)
ENDIF

cAMPM := RIGHT(cTime, 1)
IF cAMPM == 'A'
   AADD(aPlay, {nGlobalHandle, 34})      // "AM"
ELSE
   AADD(aPlay, {nGlobalHandle, 35})      // "PM"
ENDIF

nStatus := pd_PlaySeq(nChan, aPlay)

RETURN nStatus


//
FUNCTION messInfoProc()
// check status of play of message information

pd_DispMsg('messInfoProc')

IF nEvent == T_EOF .OR. nEvent == T_TERMDT
   // normal termination, play the message itself
   aCurBlock[nChan] := {    { || HearInit() },  { || HearProc() }   }
ELSE
   // hang-up
   aCurBlock[nChan] := {    { || OnHookInit() },  { || OnHookProc() }   }
ENDIF

RETURN NIL

//
FUNCTION HearInit()
// playback the recorded message

LOCAL nStatus,;
      nHandle

pd_DispMsg('HearInit')

// play the message itself
nStatus := pd_Play(nChan, aRecFile[nChan], @nHandle)

aFileHandle[nChan] := nHandle

RETURN nStatus

//
FUNCTION HearProc()
// check for successful play of message menu
pd_DispMsg('HearProc')

IF nEvent == T_EOF .OR. nEvent == T_TERMDT

   // next activity is to play the options menu.
   aCurBlock[nChan] := {    { || MessMenuInit() },  { || MessMenuProc() }   }
ELSE
   // hang-up
   aCurBlock[nChan] := {    { || OnHookInit() },  { || OnHookProc() }   }
ENDIF

// close the message itself
pd_VOXClose( aFileHandle[nChan] )

RETURN NIL

//
FUNCTION MessMenuInit()
// present the menu for message operations

LOCAL nStatus,;
      nHandle

pd_DispMsg('MessMenuInit')

// "Press 1 to review, 2 to delete, 3 to save"
nStatus := pd_Play(nChan, 'MOPTIONS.VOX', @nHandle)

aFileHandle[nChan] := nHandle

RETURN nStatus

//
FUNCTION MessMenuProc()
// check for successful play of message menu
pd_DispMsg('MessMenuProc')

DO CASE
   CASE nEvent == T_EOF .OR. nEvent == T_TERMDT
      // next activity is to play the options menu.
      aCurBlock[nChan] := {    { || OptMessInit() },  { || OptMessProc() }   }

   OTHERWISE

      // go back to the introduction
      aCurBlock[nChan] := {    { || messInfoInit() },  { || messInfoProc() }   }
ENDCASE

pd_VOXClose( aFileHandle[nChan] )

RETURN NIL

//
FUNCTION OptMessInit()
// set up for input on the message options
LOCAL nStatus

pd_DispMsg('OptMessInit')
pd_global('termdtmf', '@')
pd_global('maxsil',    5)

nStatus := pd_getDTMFs(nChan)

RETURN nStatus

//
FUNCTION OptMessProc()
LOCAL cBufContent

pd_DispMsg('OptMessProc')

IF nEvent == T_TERMDT
      // they did press _SOMETHING_

      cBufContent := pd_getBuf(nChan)
      DO CASE
         CASE cBufContent == '1'
            // repeat the message
            // go back to the message menu
            aCurBlock[nChan] := {    { || HearInit() },  { || HearProc() }   }

         CASE cBufContent == '2'
            // delete the message
            aCurBlock[nChan] := {    { || DelMessInit() },  { || DelMessProc() }   }

         CASE cBufContent == '3'
            // save the message
            aCurBlock[nChan] := {    { || SaveMessInit() },  { || SaveMessProc() }   }

         OTHERWISE
            // go back to the message menu
            aCurBlock[nChan] := {    { || MessMenuInit() },  { || MessMenuProc() }   }
      ENDCASE

ELSEIF nEvent == T_SIL
   // replay the message, maybe they fell asleep...
   aCurBlock[nChan] := {    { || messMenuInit() },  { || messMenuProc() }   }

ELSE
   aCurBlock[nChan] := {    { || OnHookInit() },  { || OnHookProc() }   }
ENDIF

RETURN NIL

//
FUNCTION DelMessInit()
// delete the current message

LOCAL nGlobalHandle,;
      nRecNo,;
      aPlay := {},;
      nStatus

pd_DispMsg('DelMessInit')

nGlobalHandle := pd_handGlobal()

// re-locate to the current record
nRecNo := FindRecord()

MESSANS->( DBGOTO(nRecNo) )
MESSANS->( DBDELETE() )                  // mark it as deleted

// set static to pack file
PackMessAns()

AADD(aPlay, {nGlobalHandle, 400})                // "Message"
AADD(aPlay, {nGlobalHandle, aCurMess[nChan]})    //  <number>
AADD(aPlay, {nGlobalHandle, 403})                // "will be deleted when u hang up"

nStatus := pd_PlaySeq(nChan, aPlay)

// add one to the current message number
aCurMess[nChan]++

// you have one less new message now
aNewMess[nChan]--

RETURN nStatus

//
FUNCTION DelMessProc()
// check for sucessful play of the delete message message
pd_DispMsg('DelMessProc')

IF nEvent == T_EOF .OR. nEvent == T_TERMDT
   IF aCurMess[nChan] <= aNewMess[nChan]
      aCurBlock[nChan] := {    { || messInfoInit() },  { || messInfoProc() }   }
   ELSE
      aCurBlock[nChan] := {    { || StartMenuInit() },  { || StartMenuProc() }   }
   ENDIF
ELSE
   aCurBlock[nChan] := {    { || MessMenuInit() },  { || MessMenuProc() }   }
ENDIF

RETURN NIL

//
STATIC FUNCTION PackMessans()
// this function packs the MESSANS database.

SELECT MESSANS
IF MESSANS->(FLOCK())
   MESSANS->(DBGOTOP())
   DO WHILE .NOT. MESSANS->(EOF())
      IF MESSANS->(DELETED())
         // delete the VOX file
         FERASE( RTRIM(MESSANS->message_id) + '.VOX' )
      ENDIF
      MESSANS->( DBSKIP() )
   ENDDO

   PACK                 // Pack MESSANS .i.e physically delete the messages
   UNLOCK
ENDIF

RETURN NIL

//
FUNCTION SaveMessInit()
// initialize function to save the message
LOCAL nGlobalHandle,;
      nRecNo,;
      nStatus,;
      aPlay             := {}                    // array of indexes to play

pd_DispMsg('SaveMessInit')

nGlobalHandle := pd_handGlobal()

nRecNo := FindRecord()

MESSANS->(DBGOTO(nRecNo))
AADD(aPlay, {nGlobalHandle, 400})                // "Message"
AADD(aPlay, {nGlobalHandle, aCurMess[nChan]})    //  <number>
AADD(aPlay, {nGlobalHandle, 404})                // "will be saved when u hang up"

nStatus := pd_PlaySeq(nChan, aPlay)

// update current message pointer
aCurMess[nChan]++

RETURN  nStatus

//
FUNCTION SaveMessProc()
// check for play of save message message

pd_DispMsg('SaveMessProc')

IF nEvent == T_EOF .OR. nEvent == T_TERMDT
   IF aCurMess[nChan] <= aNewMess[nChan]
      aCurBlock[nChan] := {    { || messInfoInit() },  { || messInfoProc() }   }
   ELSE
      aNewMess[nChan]  := 0 // There are no new messages. We are done w/ them
      aCurBlock[nChan] := {    { || StartMenuInit() },  { || StartMenuProc() }   }
   ENDIF
ELSE
   aCurBlock[nChan] := {    { || MessMenuInit() },  { || MessMenuProc() }   }
ENDIF

RETURN  NIL

// *************the rest are kind of "support" functions*********************
//
STATIC FUNCTION FindRecord()
LOCAL lStopSearch := .F.,;
      nMessage    := aCurMess[nChan],;           // set to look for current message number
      nRecNo

pd_DispMsg('FindRecord')

MESSANS->(DBGOTOP())
DO WHILE  .NOT. MESSANS->(EOF())  .AND. .NOT. lStopSearch
   // if its for this person and its new, or its a general message, choose it
   IF (aPerson[nChan] == RTRIM(MESSANS->person) .AND. MESSANS->status == "NEW") .OR.;
             aPerson[nChan] == "GENERAL"
      // Check to see if this is the current message I am looking at
      IF nMessage == aCurMess[nChan]
         // Right person and the right message number
         lStopSearch := .T.
      ELSE
         // Right person but the wrong message number
         nMessage++
      ENDIF
   ENDIF
   MESSANS->(DBSKIP())
ENDDO

MESSANS->(DBSKIP(-1))
nRecNo := MESSANS->(RECNO())

RETURN nRecNo


//
STATIC FUNCTION RemTermDTMF(cBufContent)
// remove terminating DTMF "#" character if present

IF RIGHT(cBufContent,1) == '#'
   cBufContent := LEFT(cBufContent, LEN(cBufContent) - 1 )
ENDIF

RETURN cBufContent

//
STATIC FUNCTION MilAMPM(cMilTime)
// convert HH:MM:SS military time string to AMPM format
LOCAL nHours,;
      cNewTime

   IF EMPTY(cMilTime)
      RETURN(cMilTime)
   ENDIF

   // strip seconds
   cMilTime := LEFT(cMilTime,5)
   nHours  := VAL(LEFT(cMilTime,2))

   DO CASE
      CASE nHours == 0
         // hour will be 12, minutes remain the same, append 'A'
         cNewTime := '12' + SUBSTR(cMilTime, 3) + 'A'

      CASE nHours < 12
         IF nHours < 10
            // remove leading zero, hour and minutes remain the same, append 'A'
            cNewTime := ' ' + SUBSTR(cMilTime, 2) + 'A'
         ELSE
            // hour and minutes remain the same, just add the "A"
            cNewTime := cMilTime+"A"
         ENDIF

      CASE nHours == 12
         // hour and minutes remain the same, append 'P'
         cNewTime := cMilTime + 'P'

      OTHERWISE
         // subtract 12 from hour, minutes remain the same, append 'P'
         cNewTime := STR(nHours-12, 2) + SUBSTR(cMilTime, 3) + 'P'

   ENDCASE

RETURN cNewTime


//
STATIC FUNCTION Holiday(dTestDate, aHolidays)
// see if the date is found in the array of holiday dates.
LOCAL lFound            := .F.

IF ASCAN(aHolidays, dTestDate) > 0
   lFound := .T.
ENDIF

RETURN lFound

//
STATIC FUNCTION Monday(dDate, nMonday)
// return the nth monday of the month the date is in.
LOCAL dTestDate,;
      dStartMonth

dStartMonth       := BOM(dDate)

// This finds the first monday of the month.
dTestDate := dStartMonth
DO WHILE DOW(dTestDate) <> 2
   ++dTestDate
ENDDO

RETURN(dTestDate + (nMonday -1) * 7)

//
STATIC FUNCTION LegalHolidays(dDate)
// Returns an array of all legal holidays in the year of the date passed that
// could occur on a work day.
LOCAL cYear             := STR(YEAR(dDate), 4),;
      dNewYears,;
      dTempDate,;
      aHolidays         := {}

AADD(aHolidays, (dNewYears := CTOD('01/01/' + cYear)) )   // New year's day.
AADD(aHolidays, Monday(dNewYears, 3))            // M.L. King day, 3rd Monday in Jan.
AADD(aHolidays, Monday(CTOD('02/01/' + cYear), 3))        // President's Day

// figure out the last monday in may
dTempDate := EOM(CTOD('05/01/' + cYear))
DO WHILE DOW(dTempDate) <> 2
   --dTempDate
ENDDO

AADD(aHolidays, dTempDate)                       // Memorial Day.
AADD(aHolidays, CTOD('07/04/' + cYear))          // Fourth of July
AADD(aHolidays, Monday(CTOD('09/01/' + cYear), 1)) // Labor Day, First Monday in Sept.

// find the last day of November in the current year.
dTempDate := EOM(CTOD('11/01/' + cYear))
// day five is thursday.
DO WHILE DOW(dTempDate) <> 5
   --dTempDate
ENDDO

AADD(aHolidays, CTOD('11/11/' + cYear))          // Veteran's Day
AADD(aHolidays, dTempDate)                       // Thanksgiving, last thursday of November.
AADD(aHolidays, CTOD('12/25/' + cYear))          // Christmas.

RETURN aHolidays

//
STATIC FUNCTION WinMsg(cMessage)
// display informative excuse for processing pauses
// this is really a poor excuse for a message window, but its simple!

LOCAL nLeft,;
      nRight

CLEAR SCREEN

nLeft  := MAXCOL()/2 - LEN(cMessage)/2 - 1
nRight := MAXCOL()/2 + LEN(cMessage)/2 + 2

// draw a box
@ 9, nLeft TO 11, nRight + 1 DOUBLE COLOR '+R/B'

@ 10, nLeft + 1 SAY PADC(cMessage, nRight - nLeft) COLOR 'R/B'

RETURN NIL

//
STATIC FUNCTION BOM(dDate)
// return the first date of the month of the date passed
LOCAL cDate                                      // date passed as a character

cDate := DTOC(dDate)

RETURN CTOD(LEFT(cDate, 2) + '/01/' + RIGHT(cDate, 2) )

//
STATIC FUNCTION EOM(dDate)
LOCAL dEOM

dEOM := dDate

DO WHILE MONTH(dDate) == MONTH(dEOM)
   dEOM++
ENDDO

dEOM--

RETURN dEOM

//
STATIC FUNCTION RecordSetup()

// set up parameters properly for recording
pd_reset()                                       // reset the record/playback parameters
pd_global('loopsig', 1)                          // terminate on hang-up
pd_global('maxsec',     MAX_RECORDING)           // maximum recording time
pd_global('maxsil',     MAX_REC_SILENCE)         // maximum silence
pd_global('rwbflags',         RW_TONE)           // makes a tone before recording
pd_global('rwbdata1',               5)           // duration of tone in miliseconds

RETURN NIL

//
FUNCTION ViewMessages()
// view the messages in a browse
LOCAL cScreen

SETCURSOR(SC_NORMAL)
// save the screen area
MESSANS->( BROWSE(3, 3, MAXROW()-3, MAXCOL()-3) )

SETCURSOR(SC_NONE)

RETURN NIL
