/*
   This demonstration program shows a simple telemarking application
   where the system places outbound calls on each channel based on a phone
   book database, analyzes the calls, and records the status of the call
   and the interest level of the customer in the database.

   This program uses the "state machine" concept to process events placed
   in an event queue by the Dialogic voice board.

   The program code makes reference to the "INITIATE" function and the
   "PROCESS" function.  An INITIATE and a PROCESS function are always set
   up in pairs.  First an INITIATE function will start some activity, say
   putting a channel's phone line on-hook.  (hanging it up)  Then the
   PROCESS function will check the status of the activity, and set up
   the next pair of activities.

   In this way, the flow of processing is always flexible.  Depending
   on the result of an action created by a INITIATE function, the
   PROCESS function determines what to do next.

   NOTE:  The Dialogic board hardware interrupt is defined as a manifest
          constant, HWIRQ.  Make sure you change this to be the same
          as the hardware interrupt for your Dialogic board before you
          complile pdcall.prg to try it, or always pass the irq to the
          program on the command line:

          PDCALL 5

          (set to IRQ 5)

   Compile pdcall by:

     CLIPPER pdrec /m /n /q /v /a /w /b

   This will include debugging trace information so that you can single-step
   through the program with the Clipper debugger and get a better idea of
   the flow.  Use the "code block trace" option ON so that you can see where
   code blocks were created as they are being executed.

   The DEBUGDELAY manifest constant is set to a small increment of time.
   In the state machine, if you use INKEY() rather than INKEY(nSmall),
   the state machine won't stop long enough for you to pop-up to debugger.
   DEBUGDELAY gives it enough time to see an "Alt-D" without affecting
   the performance of the system.  For a final version, you could remove
   the delay.

   DATABASE: PHNEBOOK.DBF

   Field      Field Name      Type      Length      Description
Ŀ
 1          FIRST           C         20          Customer first name       
 2          MIDDLE          C         1           Customer middle initial   
 3          LAST            C         20          Customer last name        
 4          PHONE_NUM       C         30          Phone number              
 5          COMPLETE        N         1           Call completion status    
                                                    0 - has not been called 
                                                    1 - call in progress    
                                                    2 - call complete       
                                                                            
 6          CALLS           N         8           Number of calls placed    
                                                  to this customer          
 7          INTERESTED      L                     Were they interested?     
 8          HANGUP          L                     Did they hang up?         

*/

#include "PDDIAL.CH"
#include "INKEY.CH"

#define HWIRQ         5  // hardware interrupt of Dialogic board
#define DEBUGDELAY  .05  // set to a small amount for debugger
#define NUMRINGS      1  // number of phone rings before answering

// Application defined events should be greater than T_DTMF (40)
#define T_OFFHOOK   100  // event code for automatic off hook, value should
#define T_LOCATE    101  // event code to locate the next customer
#define T_ABORT     102  // event code to shut down the channel

#define INITIATE      1  // array id for initiate code block
#define PROCESS       2  // array id for process code block

/* NOTE:  All of these STATIC arrays would be very nice to include in an
          object, such as a channel object, utilizing Class(y) or SuperClass.
          This makes coding a little less confusing and cleaner.
*/

// File wide statics
STATIC lPlayMode,;       // play back mode
       nChan,;           // channel of current event
       nEvent,;          // current event
       nEventData,;      // usually unimportant, but returned
       aCurBlock,;       // array of current initiate and process code blocks
       aFileHandle,;     // array of current file handles opened
       aRecNo,;          // array of records in PHNEBOOK.DBF for each channel
       aTerminate        // termination flags for each channel

//
FUNCTION PD_Call()
// main function containing the state machine.

LOCAL nDevType,;                      // usually unimportant, but returned
      nError,;                        // error code storage
      nMaxChan                        // maximum number of channels available

WinMsg('Processing customer calls: PHNEBOOK.DBF')

IF OpenFiles()
   // we were able to open the database sucessfully

   // Start the voice system
   IF (nError := pd_init(HWIRQ, @nMaxChan,;
                       C_LC + C_RING + C_OFFH + C_ONH, NUMRINGS)) == 0

     // Size channel array of current initiate and process code blocks
     aCurBlock   := ARRAY(nMaxChan)

     // Size channel arrays

     // NOTE: an array of channel objects would be nicer, but we're doing
     //       straight Clipper code here...
     aFileHandle := ARRAY(nMaxChan)                  // Channel file handle storage
     aRecno      := ARRAY(nMaxChan)                  // Record number pointer storage
     aTerminate  := AFILL(ARRAY(nMaxChan), .F.)      // Terminate flags

     // Set initial state of all channels
     FOR nChan := 1 TO nMaxChan
       aCurBlock[nChan] := {{ || OnHookInit() }, { || OnHookProc() }}
     NEXT

     // this is the "state machine" loop.
     DO WHILE INKEY(DEBUGDELAY) <> K_ESC .AND. ;
              ASCAN(aTerminate, {|elem| elem == .F.} ) <> 0
       // see if there are any channels left that have not been terminated

       // Did an event occur?
       IF pd_getEvent(@nChan, @nEvent, @nDevType, @nEventData) == -1

         // Process the event which will save the next channel state
         // (this is where the "process" function is executed.)
         EVAL(aCurBlock[nChan, PROCESS])

         // Initiate the new state checking the error code returned
         // (this is where the "initiate" function is executed.)
         //  EVAL(aCurBlock[nChan, INITIATE])
         IF (nError := EVAL(aCurBlock[nChan, INITIATE])) <> 0
             pd_error(nError)
         ENDIF
       ENDIF
     ENDDO

     // Stop the voice system
     pd_deinit(nMaxChan)
   ELSE
     // Display the error message
     pd_error(nError)
   ENDIF
ENDIF

RETURN NIL

//
FUNCTION OnHookInit()

// Initiate function to set the channel on hook.

LOCAL nRetVal

// Set the current channel on hook
nRetVal := pd_setHook(nChan, H_ONH)

RETURN nRetVal

//
FUNCTION OnHookProc()

// Process function to set the channel on hook.

// Proceed to wait for a ring
aCurBlock[nChan] := {{ || FindNextInit() }, { || FindNextProc() }}

RETURN NIL

//
FUNCTION FindNextInit()
// Find the next person to call in the database.

// look for the next person that has not been called
PHNEBOOK->(DBSEEK('0'))

IF PHNEBOOK->(FOUND())
   // there was someone to call!

   aRecNo[nChan] := PHNEBOOK->(RECNO())   // save the record number

   pd_putEvent(nChan, T_LOCATE, 0)  // data parm (parm 3) unimportant
ELSE
   // we ran out of people to call, terminate the state machine after
   // all calls are complete
   aTerminate[nChan] := .T.   // STATIC
   pd_putEvent(nChan, T_ABORT, 0)  // data parm (parm 3) unimportant

ENDIF

RETURN E_SUCC  // always successful

//
FUNCTION FindNextProc()

// Process function for when the channel rings.

// After we have someone to call, next we call 'em.
IF nEvent == T_LOCATE
   aCurBlock[nChan] := {{ || CallInit() }, { || CallProc() }}

ELSEIF nEvent == T_ABORT
   // set code blocks to do nothing until all channels are shut down
   aCurBlock[nCHan] := {{|| E_SUCC }, {|| NIL } }
ELSE
   // Unexpected event detected, must wait for another ring
   aCurBlock[nChan] := {{ || OnHookInit() }, { || OnHookProc() }}
ENDIF

RETURN NIL

//
FUNCTION CallInit()

// Initiate function to place a call.

LOCAL nRetVal,;
      aStatus


// call the number stored in the database record, locate to the proper record
// in the database
PHNEBOOK->(DBGOTO(aRecNo[nChan]))

IF PHNEBOOK->(RLOCK())
   // if the record lock is successful...
   REPLACE PHNEBOOK->complete WITH 1   // 1 status indicates call in progress
                                       // this keeps the other channels from
                                       // calling the same guy.
ENDIF

nRetVal := pd_callp(nChan, RTRIM(PHNEBOOK->phone_num))

//  get the call status information
pd_GetCStat(nChan, aStatus)

RETURN nRetVal  // always successful

//
FUNCTION CallProc()

// Process function after a call has been initiated.

// If the call event was terminated, see if it was a connect...
IF nEvent == T_CATERM
   IF nEventData == CA_CONN
      // connect message, can also check for CA_BUSY, CA_NOAN, CA_NORNG,
      // CA_OPINT.  Using these codes could make processing more sophisticated,
      // so that it could, for example, log a user as dead if CA_NORNG
      // (no ring) was received.  A busy could retry real soon, and a
      // no answer would wait a while before re-trying.
      aCurBlock[nChan] := {{ || IntroInit() }, { || IntroProc() }}
   ELSE
      // go to the person's record and update the database with the
      // fact that we called them once again.
      PHNEBOOK->(DBGOTO(aRecno[nChan]))
      IF PHNEBOOK->(RLOCK())
         REPLACE PHNEBOOK->calls    WITH PHNEBOOK->calls + 1,;
                 PHNEBOOK->complete WITH 0
      ENDIF

      // go back on-hook
      aCurBlock[nChan] := {{ || OnHookInit() }, { || OnHookProc() }}
   ENDIF
ELSE
   PHNEBOOK->(DBGOTO(aRecno[nChan]))
   IF PHNEBOOK->(RLOCK())
      REPLACE PHNEBOOK->calls    WITH PHNEBOOK->calls + 1,;
              PHNEBOOK->complete WITH 0
   ENDIF
   // Unexpected event detected, must wait for another ring
   aCurBlock[nChan] := {{ || OnHookInit() }, { || OnHookProc() }}
ENDIF

RETURN NIL

//
FUNCTION IntroInit()

// Initiate function to handle playing the introductory message.

LOCAL nHandle,;
      nRetVal

// This is a good point to set the parameters for this message play.
// Things like terminating character, etc. e.g.
pd_global('termdtmf', '@')   // any key will terminate message play

// Play the file, save the file handle so it can be closed later.
// (by the process function)

/* file contains:
   Hi, I'm calling from mega-mortgage! Home re-financing is now available at
   five percent APR with zero points!  At this fantastic interest rate, would
   you be interested in re-financing your home?  If you would, press the one
   key on your telephone.  IF you are not interested in re-financing at this
   time, press the two key on your telephone.
*/
nRetVal := pd_play(nChan, 'SURVEY.VOX', @nHandle)
aFileHandle[nChan] := nHandle

RETURN nRetVal

//
FUNCTION IntroProc()

// Process function after the introductory message has played.

IF nEvent == T_EOF .OR. nEvent == T_TERMDT
  // Check here for other states also.
  aCurBlock[nChan] := {{ || InputInit() }, { || InputProc() }}

ELSE
  // hang-up next if the file was abnormally played.
  aCurBlock[nChan] := {{ || OnHookInit() }, { || OnHookProc() }}

ENDIF

// Close the most recently opened file
pd_VOXClose(aFileHandle[nChan])

RETURN NIL

//
FUNCTION InputInit()
// initialize for user input
LOCAL nRetVal

/* NOTE: Because the buffer is not cleared here (it could be), if a choice
         is made during the play of the announcement, it will already be in
         the buffer and InputProc will see it.  If you don't want this,
         do a pd_ClrDTMF() here!  */

nRetVal := pd_getDTMFs(nChan)

RETURN nRetVal

//
FUNCTION InputProc()
// accept input for menu selection and process the option.
LOCAL cDtmfBuf,;
      nPos

// Check for terminating dtmf
IF nEvent == T_TERMDT
   cDtmfBuf := pd_getBuf(nChan)

   PHNEBOOK->(DBGOTO(aRecNo[nChan]))
   // choice 1 or 2
   DO CASE
      CASE cDTMFBuf == '1'
         // they liked it!
         IF PHNEBOOK->(RLOCK())
            REPLACE PHNEBOOK->interested WITH .T.,;
                    PHNEBOOK->complete   WITH 2    // mode 2 indicates they are complete
         ENDIF
         // This would be a primo place to transfer the caller to a live
         // person!!

         // say "bye bye"  Paul McCartney
         aCurBlock[nChan] := {{ || GByeInit() }, { || GByeProc() }}

      CASE cDTMFBuf == '2'
         // they are not interested
         IF PHNEBOOK->(RLOCK())
            REPLACE PHNEBOOK->interested WITH .F.,;
                    PHNEBOOK->complete   WITH 2
         ENDIF
         aCurBlock[nChan] := {{ || GByeInit() }, { || GByeProc() }}

      OTHERWISE
         // they are not smart enough to run the telephone, or they
         // have pulse dialing.
         IF PHNEBOOK->(RLOCK())
            REPLACE PHNEBOOK->interested WITH .F.,;
                    PHNEBOOK->complete   WITH 2,;
                    PHNEBOOK->hangup     WITH .T.
         ENDIF

         aCurBlock[nChan] := {{ || OnHookInit() }, { || OnHookProc() } }
   ENDCASE
ENDIF

RETURN NIL

//
FUNCTION GByeInit()

// Initiate function to handle playing the goodbye message.

LOCAL nHandle,;
      nRetVal

// Play the goodbye message

// Play the file, save the file handle to close it later
nRetVal := pd_play(nChan, 'PDGBYE.VOX', @nHandle)
aFileHandle[nChan] := nHandle

RETURN nRetVal

//
FUNCTION GByeProc()
// Process function to handle playing the goodbye message.

// close the file
pd_VOXClose(aFileHandle[nChan])

// Put channel back on hook
aCurBlock[nChan] := {{ || OnHookInit() }, { || OnHookProc() }}

RETURN NIL

//
FUNCTION OpenFiles()
// open any DBFs required for this application
LOCAL lSuccess          := .T.

IF FILE('PHNEBOOK.DBF')
   // open the file shared
   USE PHNEBOOK SHARED NEW
   IF FILE('PHNEBOOK.NTX')
      // if the index file exists, open it!
      PHNEBOOK->(DBSETINDEX('PHNEBOOK.NTX'))
   ELSE
      // build the index if it does not exist.
      PHNEBOOK->(DBCREATEINDEX('PHNEBOOK.NTX',;
                                STR(PHNEBOOK->complete, 1) + PHNEBOOK->phone_num,;
                                {|| STR(PHNEBOOK->complete, 1) + PHNEBOOK->phone_num},;
                                .T.))
   ENDIF
ELSE
   ALERT("File PHNEBOOK.DBF could not be found; it's required.")
   lSuccess := .F.
ENDIF

RETURN lSuccess

//
FUNCTION WinMsg(cMessage)
// display informative excuse for processing pauses
// this is really a poor excuse for a message window, but its simple!

LOCAL nLeft,;
      nRight

CLEAR SCREEN

nLeft  := MAXCOL()/2 - LEN(cMessage)/2 - 1
nRight := MAXCOL()/2 + LEN(cMessage)/2 + 2

// draw a box
@ 9, nLeft TO 11, nRight + 1 DOUBLE COLOR '+R/B'

@ 10, nLeft + 1 SAY PADC(cMessage, nRight - nLeft) COLOR 'R/B'

RETURN NIL


