/*
   This demonstration program presents a menu to the caller.  Pressing
   a certain menu option will cause a particular song to play.

   This code makes reference to the "INITIATE" function and the "PROCESS"
   function.  An INITIATE and PROCESS function are always set up in
   pairs.  First an INITIATE function will start some activity, say
   putting a channel's phone line on-hook.  (hanging it up)  Then the
   PROCESS function will check the status of the activity, and set up
   the next pair of activities.

   In this way, the flow of processing is always flexible.  Depending
   on the result of an action created by a INITIATE function, the
   PROCESS function determines what to do next.

   NOTE:  The Dialogic board hardware interrupt is defined as a manifest
          constant, HWIRQ.  Make sure you change this to be the same
          as the hardware interrupt for your Dialogic board before you
          complile pdmenu.prg to try it, or always pass the irq to the
          program on the command line.

   Compile pdmenu by:

     CLIPPER pdmenu /m /n /q /v /a /w /b

   This will include debugging trace information so that you can single-step
   through the program with the Clipper debugger and get a better idea of
   the flow.  Use the "code block trace" option ON so that you can see where
   code blocks were created as they are being executed.

   The DEBUGDELAY manifest constant is set to a small increment of time.
   In the state machine, if you use INKEY() rather than INKEY(nSmall),
   the state machine won't stop long enough for you to pop-up to debugger.
   DEBUGDELAY gives it enough time to see an "Alt-D" without affecting
   the performance of the system.  For a final version, you could remove
   the delay.

*/

#include "PDDIAL.CH"
#include "INKEY.CH"

#define HWIRQ         5  // hardware interrupt of Dialogic board
#define DEBUGDELAY  .05  // set to a small amount for debugger
#define MAXCHAN       4  // maximum number of channels
#define NUMRINGS      1  // number of phone rings before answering
#define TERMDTMF    '#'  // standard terminating DTMF character
#define MNUMRINGS     1  // number of rings to wait before answering

#define INITIATE      1  // array id for initiate code block
#define PROCESS       2  // array id for process code block

// File wide statics
STATIC lStateLoop,;      // terminates execution of the state machine
       nChan,;           // channel of current event
       nEvent,;          // current event
       aCurBlock,;       // array of current initiate and process code blocks
       aFileHandle       // array of current file handles opened

//
FUNCTION StartUp(cHWIRQ)
// initialize the state machine processing.
// can pass the hardware interrupt number from the DOS command line

LOCAL nChannel,;                                 // loop counter for initialization of the channels
      nDevType,;                                 // device type on which the error occured
      nEventData,;                               // data variable passed back with the event code
      nErr,;                                     // init error code
      nHWIRQ,;                                   // hardware interrupt for board
      nInKey,;                                   // key input value
      nMaxChan          := 0                     // maximum number of channels

// If no hardware interrupt was passed as a parameter, set the hardware
// interrupt value to the one defined by the manifest constant.
IF EMPTY(cHWIRQ)
   nHWIRQ := HWIRQ
ELSE
   // convert the hardware interrupt to a numeric.
   nHWIRQ := VAL(cHWIRQ)
ENDIF

CLEAR SCREEN

// Start the voice system
IF (nErr := pd_init(nHWIRQ, @nMaxChan,;
                    C_LC + C_RING + C_OFFH + C_ONH, MNUMRINGS)) == 0
  // there was no error doing the initialization of the driver, etc.

  // Set initial state: re-initialize channel array and file handle arrays.
  // NOTE TO THOSE WITH "CLASS": A voice channel is a good candidate for
  //                             an class object!!

  aCurBlock   := {}
  aFileHandle := {}

  FOR nChannel := 1 TO nMaxChan
      // One array element per channel.  Each array element contains
      // a two element array containing the next initiate function and
      // the next process function.  All channels start on-hook.
      AADD(aCurBlock,  {{ || OnHookInit() }, { || OnHookProc() }})

      // a storage space to keep the channel's file handle
      AADD(aFileHandle, 0)

  NEXT

  // set globals to start with
  pd_global('loopsig', 1)                        // generate event on dropped line
  pd_global('termdtmf', '@')                     // terminate on any character

  // Loop to call the voice handler
  // processing events with the state machine

  // yo, this is the "state machine" processing loop>>>>>
  DO WHILE (nInKey := INKEY(DEBUGDELAY)) <> K_ESC
    

    // Did an event occur?
    IF pd_getEvent(@nChan, @nEvent, @nDevType, @nEventData) == -1

      // Placing this message in here slows us down a little.  It's
      // certainly not necessary.  Its just an example of how you can
      // put your own function calls which can do other types of
      // processing inside the state machine loop!
      WinMsg('Processing Events, ESC to exit')

      // Process the event and save the next channel state
      // (execute the "PROCESS" function)
      EVAL(aCurBlock[nChan, PROCESS])

      // execute the INITIATE function.  (this sets the code blocks
      // defining the next activity to execute.)

      IF (nErr := EVAL(aCurBlock[nChan, INITIATE])) <> 0
         // process the error message
         pd_error(nErr)
      ENDIF
    ENDIF
  ENDDO

  // Shut down the voice system
  IF pd_deinit(nMaxChan) <> 0
    WinMsg( ALLTRIM(pd_d4xerr(nErr)) )
  ENDIF
ELSE
  WinMsg(  ALLTRIM(pd_d4xerr(nErr)) )
ENDIF


RETURN NIL

//
FUNCTION OnHookInit()
// Initiate function to set the channel on hook.
LOCAL nRetVal

pd_dispMsg('OnHookInit')

// Set the current channel on hook
nRetVal := pd_setHook(nChan, H_ONH)

RETURN nRetVal

//
FUNCTION OnHookProc()
// Process function to set the channel on hook.
pd_dispMsg('OnHookProc')

// Proceed to wait for a ring
aCurBlock[nChan] := {{ || RingInit() }, { || RingProc() }}

RETURN NIL

//
FUNCTION RingInit()
// Initiate function to wait for a channel to ring.
pd_dispMsg('RingInit')

// Clear the dtmf buffer for the channel and wait for a call.
pd_clrDTMF(nChan)

RETURN E_SUCC  // always successful

//
FUNCTION RingProc()
// Process function for when the channel rings.
pd_dispMsg('RingProc')

// When a call is detected, set state to go off hook
IF nEvent == T_RING
  // Proceed to take the channel off hook
  aCurBlock[nChan] := {{ || OffHookInit() }, { || OffHookProc() }}
ELSE
  // Unexpected event detected, must wait for another ring
  aCurBlock[nChan] := {{ || RingInit() }, { || RingProc() }}
ENDIF

RETURN NIL

//
FUNCTION OffHookInit()
// Initiate function to take the channel off hook.
LOCAL nRetVal

pd_dispMsg('OffHookInit')


// Channel activity detected, go off hook
nRetVal := pd_setHook(nChan, H_OFFH)

RETURN nRetVal

//
FUNCTION OffHookProc()
// Process function to determine if the channel was answered and stay off hook.
pd_dispMsg('OffHookProc')

// When channel goes off hook proceed
IF nEvent == T_OFFH
   aCurBlock[nChan] := {{ || IntroInit() }, { || IntroProc() }}
ELSE
   // Unexpected event detected, must go back on hook
   aCurBlock[nChan] := {{ || OnHookInit() }, { || OnHookProc() }}
ENDIF

RETURN NIL

//
FUNCTION IntroInit()
// Initiate function to handle playing the menu message.

LOCAL nHandle,;
      nRetVal

pd_dispMsg('IntroInit')

// This is a good point to set the parameters for this message play.
// Things like terminating character, etc. e.g.
// pd_global('termdtmf', '#')   // # key will terminate message play

/* If you have a menu that has more options than you have touch-tone keys,
   you will have to set termdtmf to be '#' so that you can enter a menu
   selection like "13#" to do menu option number 13.  If you have termdtmf
   set to be "@" for quick menu selection, don't forget to set it back to
   "#" if you are doing DTMF data entry later!!!  */

// Play the file, save the file handle so it can be closed later.
// (by the process function)
nRetVal := pd_play(nChan, 'PDMENU.VOX', @nHandle)

aFileHandle[nChan] := nHandle                    // save the handle

RETURN nRetVal

//
FUNCTION IntroProc()
// Process function after the menu message has played.

pd_dispMsg('IntroProc')

IF nEvent == T_EOF .OR. nEvent == T_TERMDT   // here check for other states
  // normal file play completion, go into menu input process.
  aCurBlock[nChan] := {{ || MenuInputInit() }, { || MenuInputProc() }}

ELSE
  // hang-up next of the file was abnormally played.
  aCurBlock[nChan] := {{ || OnHookInit() }, { || OnHookProc() }}

ENDIF

// Close the most recently opened file
pd_VOXClose(aFileHandle[nChan])

RETURN NIL

//
FUNCTION MenuInputInit()
// initialize for menu selection
LOCAL nRetVal

pd_dispMsg('MenuInputInit')

nRetVal := pd_getDTMFs(nChan)

/* NOTE: Because the buffer is not cleared here (it could be), if a menu
         selection is made during the play of the menu announcement, it will
         already be in the buffer and MenuInputProc will see it.  If you
         don't want this, do a pd_ClrDTMF() here!  */


RETURN nRetVal

//
FUNCTION MenuInputProc()
// accept input for menu selection and process the option.
LOCAL cDtmfBuf,;
      nPos
pd_dispMsg('MenuInputProc')

// Check for terminating dtmf
IF nEvent == T_TERMDT
   cDtmfBuf := pd_getBuf(nChan)

   // pound key hangs up.
   IF cDtmfBuf == '#'
      aCurBlock[nChan] := {{ || GByeInit() }, { || GByeProc() }}
   ELSE
      cDTMFBuf := RemTermDtmf(cDtmfBuf)
      DO CASE
         CASE cDTMFBuf == '1'
            // play Gershwin
            aCurBlock[nChan] := {{ || PlayMusicInit('GERSHWIN.VOX') }, { || PlayMusicProc() }}

         CASE cDTMFBuf == '2'
            // play the CARS
            aCurBlock[nChan] := {{ || PlayMusicInit('CARS.VOX') }, { || PlayMusicProc() }}

         CASE cDTMFBuf == '3'
            // play CHILLI PEPPERS
            aCurBlock[nChan] := {{ || PlayMusicInit('CP.VOX') }, { || PlayMusicProc() }}

      OTHERWISE
         // Good place to play a message indicating that the input chosen
         // was bogus!
         aCurBlock[nChan] := {{ || IntroInit() }, { || IntroProc() }}
      ENDCASE
   ENDIF
ELSE
   // Good place to somehow log that an unexpected event was received!
   // Then hang up.
   aCurBlock[nChan] := {{ || OnHookInit() }, { || OnHookProc() }}
ENDIF

RETURN NIL

//
FUNCTION PlayMusicInit(cFile)
// Initiate function to handle playing back the recorded message.

LOCAL nHandle,;
      nRetVal

pd_dispMsg('PlayMusicInit(' + cFile + ')' )

// Play the file
nRetVal := pd_play(nChan, cFile, @nHandle)

aFileHandle[nChan] := nHandle

RETURN nRetVal

//
FUNCTION PlayMusicProc()
// Process function after playing back the recorded music.
pd_dispMsg('PlayMusicProc')

// Close the file we played
pd_VOXClose(aFileHandle[nChan])

// Go back to the menu
aCurBlock[nChan] := {{ || IntroInit() }, { || IntroProc() }}

RETURN NIL

//
FUNCTION GByeInit()

// Initiate function to handle playing the goodbye message.

LOCAL nHandle,;
      nRetVal

// Play the goodbye message
pd_dispMsg('Play PDGBYE.VOX')

// Play the file, save the file handle to close it later
nRetVal := pd_play(nChan, 'PDGBYE.VOX', @nHandle)
aFileHandle[nChan] := nHandle

RETURN nRetVal

//
FUNCTION GByeProc()
// Process function to handle playing the goodbye message.

// Close the most recently played file
pd_dispMsg('Close PDGBYE.VOX')

// close the file
pd_VOXClose(aFileHandle[nChan])

// Put channel back on hook
aCurBlock[nChan] := {{ || OnHookInit() }, { || OnHookProc() }}

RETURN NIL

//
FUNCTION WinMsg(cMessage)
// display informative excuse for processing pauses
// this is really a poor excuse for a message window, but its simple!

LOCAL nLeft,;
      nRight

CLEAR SCREEN

nLeft  := MAXCOL()/2 - LEN(cMessage)/2 - 1
nRight := MAXCOL()/2 + LEN(cMessage)/2 + 2

// draw a box
@ 9, nLeft TO 11, nRight + 1 DOUBLE COLOR '+R/B'

@ 10, nLeft + 1 SAY PADC(cMessage, nRight - nLeft) COLOR 'R/B'

RETURN NIL

//
STATIC FUNCTION RemTermDtmf(cDtmfBuf)
// Removes the terminating character DTMF character if it is the standard
// DTMF terminating character.

// Ensure rightmost character is terminating character
IF RIGHT(cDtmfBuf, 1) == TERMDTMF
  cDtmfBuf := LEFT(cDtmfBuf, LEN(cDtmfBuf) - 1)
ENDIF

RETURN cDtmfBuf

