/*
   This demonstration program shows a Clipper input screen where
   a hot key (F2) is used to record a message with a playback option.

   In this simple case, the Dialogic System and "state machine"
   processing of Dialogic events does not begin until the F2 key is
   pressed.  In a more complex application, the state machine is
   usually running all the time.

   This code makes reference to the "initiate" function and the "process"
   function.  An initiate and process function are always set up in
   pairs.  First an initiate function will start some activity, say
   putting a channel's phone line on-hook.  (hanging it up)  Then the
   process function will check the status of the activity, and set up
   the next pair of activities.

   In this way, the flow of processing is always flexible.  Depending
   on the result of an action created by a initiate function, the
   process function determines what to do next.

   NOTE:  The Dialogic board hardware interrupt is defined as a manifest
          constant, HWIRQ.  Make sure you change this to be the same
          as the hardware interrupt for your Dialogic board before you
          complile pd_rec.prg to try it.

   Compile pdrec.prg by:

     CLIPPER pdrec /m /n /q /v /a /w /b

   This will include debugging trace information so that you can single-step
   through the program with the Clipper debugger and get a better idea of
   the flow.  Use the "code block trace" option ON so that you can see where
   code blocks were created as they are being executed.

   The DEBUGDELAY manifest constant is set to a small increment of time.
   In the state machine, if you use INKEY() rather than INKEY(nSmall),
   the state machine won't stop long enough for you to pop-up to debugger.
   DEBUGDELAY gives it enough time to see an "Alt-D" without affecting
   the performance of the system.  For a final version, you could remove
   the delay.

*/

#include "PDDIAL.CH"
#include "INKEY.CH"

#define HWIRQ         5  // hardware interrupt of Dialogic board
#define DEBUGDELAY  .05  // set to a small amount for debugger
#define MAXCHAN       4  // maximum number of channels
#define NUMRINGS      1  // number of phone rings before answering

#define T_OFFHOOK   100  // event code for automatic off hook, value should
                         // be greater than T_DTMF (40)

#define INITIATE      1  // array id for initiate code block
#define PROCESS       2  // array id for process code block

// File wide statics
STATIC cRecFile,;        // name of file to record to
       lAutoMode,;       // auto answer mode
       lPlayMode,;       // play back mode
       lStateLoop,;      // terminates execution of the state machine
       nChan,;           // channel of current event
       nEvent,;          // current event
       aCurBlock,;       // array of current initiate and process code blocks
       aFileHandle       // array of current file handles opened

//
FUNCTION Main()

// Front end function to demonstrate an input screen initating the recording
// of a message via the Dialogic board.  Message is recorded in file
// MESSAGE1.VOX.

LOCAL cVar              := SPACE(10),;
      getlist           := {}

// Clear the screen and set F2 to record a message
SET SCOREBOARD OFF
CLS
SETKEY(K_F2, { || RecMsg('MESSAGE1.VOX', .F., .F.) })

// Simple input screen
DISPBOX(0, 0, 6, 79)

@1, 28 SAY 'Dialogic Demonstration'
@5,  1 SAY 'F2-Record a message after the phone rings'
@3, 20 SAY 'Input field' GET cVar
READ

// Restore F2
SETKEY(K_F2, NIL)

RETURN NIL

//
FUNCTION RecMsg(cFile, lAutoAnswer, lPlayBack)

// Initializes the Dialogic voice system and implements the "state machine" to
// record a message.
// cFile       == name of file to record to
// lAutoAnswer == .T. means the Dialogic board will answer when the phone
//                rings
// lAutoAnser  == .F. means the Dialogic board will assume the phone is
//                already answered and automatically goes off hook independent
//                of a ring
// lPlayBack   == .F. means do not play back the recorded file, .T. means play
//                back file


LOCAL nDevType,;                      // usually unimportant, but returned
      nErr,;
      nEventData,;                    // usually unimportant, but returned
      nMaxChan          := MAXCHAN

// Validate parameters
IF VALTYPE(cFile) <> 'C'
  cFile := 'MESSAGE1.VOX'
ENDIF
cRecFile := cFile

IF VALTYPE(lAutoAnswer) <> 'L'
  lAutoAnswer := .T.
ENDIF
lAutoMode := lAutoAnswer

IF VALTYPE(lPlayBack) <> 'L'
  lPlayBack := .F.
ENDIF
lPlayMode := lPlayBack

// Start the voice system
IF (nErr := pd_init(HWIRQ, @nMaxChan,;
                    C_LC + C_RING + C_OFFH + C_ONH, NUMRINGS)) == 0

  // Size channel array of current initiate and process code blocks
  aCurBlock   := ARRAY(nMaxChan)

  // Size channel array of file handles
  aFileHandle := ARRAY(nMaxChan)

  // Set initial state of all channels
  FOR nChan := 1 TO nMaxChan
    aCurBlock[nChan] := {{ || OnHookInit() }, { || OnHookProc() }}
  NEXT

  // Put an event to automatically go offhook with the assumption the phone
  // has already been answered, event is for channel 1 only
  IF .NOT. lAutoAnswer
    pd_putEvent(1, T_OFFHOOK, 0)  // data parm (parm 3) unimportant
    // move to screen row eight, column zero.
    SETPOS(8, 0)
    pd_dispMsg('Channel 1 going off hook')
  ENDIF

  // Loop to call the voice event handler
  lStateLoop := .T.

  // this is the "state machine" loop.
  DO WHILE INKEY(DEBUGDELAY) <> K_ESC .AND. lStateLoop
    // Did an event occur?
    IF pd_getEvent(@nChan, @nEvent, @nDevType, @nEventData) == -1

      // Process the event which will save the next channel state
      // (this is where the "process" function is executed.)
      EVAL(aCurBlock[nChan, PROCESS])

      // Initiate the new state checking the error code returned
      // (this is where the "initiate" function is executed.)
      IF (nErr := EVAL(aCurBlock[nChan, INITIATE])) <> 0
        pd_error(nErr)
      ENDIF
    ENDIF
  ENDDO

  // Stop the voice system
  pd_deinit(MAXCHAN)
ELSE
  // Display the error message
  pd_error(nErr)
ENDIF

RETURN NIL

//
FUNCTION OnHookInit()

// Initiate function to set the channel on hook.

LOCAL nRetVal

// Set the current channel on hook
nRetVal := pd_setHook(nChan, H_ONH)

// Stop the state machine
lStateLoop := lAutoMode

RETURN nRetVal

//
FUNCTION OnHookProc()

// Process function to set the channel on hook.

// Proceed to wait for a ring
aCurBlock[nChan] := {{ || RingInit() }, { || RingProc() }}

RETURN NIL

//
FUNCTION RingInit()

// Initiate function to wait for a channel to ring.

// Clear the dtmf buffer for the channel and wait for a call.
pd_clrDTMF(nChan)

RETURN E_SUCC  // always successful

//
FUNCTION RingProc()

// Process function for when the channel rings.

// When a call is detected, set state to go off hook
IF nEvent == T_RING .OR. nEvent == T_OFFHOOK
  // Proceed to take the channel off hook
  aCurBlock[nChan] := {{ || OffHookInit() }, { || OffHookProc() }}
ELSE
  // Unexpected event detected, must wait for another ring
  aCurBlock[nChan] := {{ || RingInit() }, { || RingProc() }}
ENDIF

RETURN NIL

//
FUNCTION OffHookInit()

// Initiate function to take the channel off hook.

LOCAL nRetVal

// Channel activity detected, go off hook
nRetVal := pd_setHook(nChan, H_OFFH)

RETURN nRetVal

//
FUNCTION OffHookProc()

// Process function to determine if the channel was answered and stay off hook.

// When channel goes off hook proceed
IF nEvent == T_OFFH
  IF lPlayMode
    aCurBlock[nChan] := {{ || PlayBackInit() }, { || PlayBackProc() }}
  ELSE
    aCurBlock[nChan] := {{ || IntroInit() }, { || IntroProc() }}
  ENDIF
ELSE
  // Unexpected event detected, must go back on hook
  aCurBlock[nChan] := {{ || OnHookInit() }, { || OnHookInit() }}
ENDIF

RETURN NIL

//
FUNCTION IntroInit()

// Initiate function to handle playing the introductory message.

LOCAL nHandle,;
      nRetVal

// Clear the global values for play, default values OK
pd_reset()

// This is a good point to set the parameters for this message play.
// Things like terminating character, etc. e.g.
pd_global('termdtmf', '@')   // any key will terminate message play

/* If you are doing a menu, and the menu has more options than you have
   touch-tone keys, you probably will have to have them terminate their
   entry by pressing the pound key.  For example, "13#" to select menu
   option 13.  If you are later going to get DTMF input of data, don't
   forget to set pd_global back to "#" or whatever, otherwise it will
   always terminate on the first key they press!  */


// Play the file, save the file handle so it can be closed later.
// (by the process function)
nRetVal := pd_play(nChan, 'INTRO.VOX', @nHandle)
aFileHandle[nChan] := nHandle

RETURN nRetVal

//
FUNCTION IntroProc()

// Process function after the introductory message has played.

IF nEvent == T_EOF           // here check for other states, T_TERMDT, etc.
  // normal file play completion, go into record next.
  aCurBlock[nChan] := {{ || RecInit() }, { || RecProc() }}

ELSE
  // hang-up next of the file was abnormally played.
  aCurBlock[nChan] := {{ || OnHookInit() }, { || OnHookProc() }}

ENDIF

// Close the most recently opened file
pd_VOXClose(aFileHandle[nChan])

RETURN NIL

//
FUNCTION RecInit()

// Initiate function to record a message.

LOCAL nHandle,;
      nRetVal

// Clear the global values for record
pd_reset()

// Set global values.
pd_global('maxsec',       120)    // maximum 120 second message
pd_global('termdtmf',     '@')    // terminate on any dtmf
pd_global('maxsil',         5)    // terminate after 5 seconds of silence
pd_global('rwbflags', RW_TONE)    // enable beep before record
pd_global('rwbdata1',       3)    // .6 second beep

nRetVal := pd_recFile(nChan, cRecFile, @nHandle, RM_SCOMP, .F.)
aFileHandle[nChan] := nHandle

RETURN nRetVal

//
FUNCTION RecProc()

// Process function after recording a message.

// When record is complete
IF nEvent == T_TERMDT
  pd_clrDTMF(nChan)
  aCurBlock[nChan] := {{ || PlayBackInit() }, { || PlayBackProc() }}
ELSE
  aCurBlock[nChan] := {{ || OnHookInit() }, { || OnHookProc() }}
ENDIF

// Close the most recently opened file
pd_VOXClose(aFileHandle[nChan])

RETURN NIL

//
FUNCTION PlayBackInit()

// Initiate function to handle playing back the recorded message.

LOCAL nHandle,;
      nRetVal

// Clear the global values for play
pd_reset()

// Set terminating character
pd_global('termdtmf', '@')    // terminate on any dtmf

// Play the file
nRetVal := pd_play(nChan, cRecFile, @nHandle)
aFileHandle[nChan] := nHandle

RETURN nRetVal

//
FUNCTION PlayBackProc()

// Process function after playing back the recorded message.

// Go back on hook
aCurBlock[nChan] := {{ || OnHookInit() }, { || OnHookProc() }}

// Close the most recently opened file
pd_VOXClose(aFileHandle[nChan])

RETURN NIL
